"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.tracePropAccess = exports.PathPartType = void 0;
var PathPartType;
(function (PathPartType) {
    PathPartType["function"] = "function";
    PathPartType["object"] = "object";
    PathPartType["string"] = "string";
    PathPartType["number"] = "number";
    PathPartType["boolean"] = "boolean";
    PathPartType["symbol"] = "symbol";
    PathPartType["undefined"] = "undefined";
    PathPartType["bigint"] = "bigint";
})(PathPartType = exports.PathPartType || (exports.PathPartType = {}));
// cleans the stack and drops last frame (ourselves)
const cleanStack = (stack) => stack.split('\n').map(line => line.trim()).filter(line => line.startsWith('at')).slice(1).join('\n');
const isDataObject = (obj) => Buffer.isBuffer(obj) || typeof obj.byteLength === 'number';
function tracePropAccess(obj, options, paths = [[]]) {
    const actualOptions = {
        callback: () => { },
        shouldFollow: () => true,
        ...options,
    };
    return new Proxy(obj, {
        get(target, propKey) {
            const reflectedProp = Reflect.get(target, propKey);
            if (!(propKey in target) /*&& propKey !== 'then'*/ ||
                !actualOptions.shouldFollow(target, propKey)) {
                const pathsSoFar = paths.slice(0, -1);
                if (pathsSoFar.length > 0) {
                    actualOptions.callback(pathsSoFar, target);
                }
                return reflectedProp;
            }
            const workingPaths = paths.map(pathArr => [...pathArr]);
            const stackTraceError = {};
            Error.captureStackTrace(stackTraceError);
            const stackTrace = cleanStack(stackTraceError.stack);
            const newPathEntry = typeof reflectedProp === 'function'
                ? {
                    key: propKey.toString(),
                    type: PathPartType.function,
                    callArgs: [],
                    stackTrace,
                }
                : {
                    key: propKey.toString(),
                    type: PathPartType[typeof reflectedProp],
                    stackTrace,
                };
            workingPaths[workingPaths.length - 1] = [
                ...workingPaths[workingPaths.length - 1],
                newPathEntry,
            ];
            if (reflectedProp) {
                if (typeof reflectedProp === 'object' && !isDataObject(reflectedProp)) {
                    return tracePropAccess(reflectedProp, actualOptions, workingPaths);
                }
                if (typeof reflectedProp === 'function') {
                    return function (...args) {
                        workingPaths[workingPaths.length - 1].pop();
                        workingPaths[workingPaths.length - 1].push({
                            key: propKey.toString(),
                            type: PathPartType.function,
                            callArgs: args,
                            stackTrace,
                        });
                        const newPaths = [...workingPaths, []];
                        const trapAsyncMethodPromise = actualOptions.maybeCreateAsyncTrap ? actualOptions.maybeCreateAsyncTrap(target, propKey, newPaths) : null;
                        const fnResult = trapAsyncMethodPromise ? trapAsyncMethodPromise.then(() => reflectedProp.apply(target, args), () => reflectedProp.apply(target, args)) : reflectedProp.apply(target, args);
                        const finishFunctionTracing = (functionResult) => {
                            newPaths.pop();
                            actualOptions.callback(newPaths, functionResult);
                        };
                        if (typeof fnResult !== 'object' || !fnResult) {
                            finishFunctionTracing(fnResult);
                            return fnResult;
                        }
                        if (typeof fnResult.then === 'function') {
                            return fnResult.catch((error) => {
                                finishFunctionTracing(error);
                                return Promise.reject(error);
                            }).then((result) => {
                                if (typeof result === 'object' &&
                                    result &&
                                    !isDataObject(result)) {
                                    return tracePropAccess(result, actualOptions, newPaths);
                                }
                                finishFunctionTracing(result);
                                return result;
                            });
                        }
                        return tracePropAccess(fnResult, actualOptions, newPaths);
                    };
                }
            }
            actualOptions.callback(workingPaths, target[propKey]);
            return reflectedProp;
        },
    });
}
exports.tracePropAccess = tracePropAccess;
